%{

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "intern.h"

int lc_cfg_i_nline=0;
extern int yylex();

typedef struct {
	int from, to;
} numericrange;

numericrange years = { 0, 0 };
char *cls=NULL;

int lc_i_cc(char **);

%}

/*
 * Token value definition
 */
%union {
	char *tv_char;
	char **tv_chars;
	double tv_double;
	int	tv_int;
	numericrange tv_nr;
	lc_t_cfdata tv_cfd;
};

/*
 * Token types returned by scanner
 */
%token	<tv_char>	TOK_STRING
%token	<tv_double>	TOK_DOUBLE
%token	<tv_int>	TOK_INT
%token	<tv_int>	TOK_MONTH
%token	<tv_int>	TOK_WDAY

%type	<tv_nr>		NumericRange
%type	<tv_nr>		HourRange
%type	<tv_nr>		YearRange
%type	<tv_cfd>	Period
%type	<tv_cfd>	HourPeriod
%type	<tv_cfd>	DayPeriod
%type	<tv_cfd>	Cost
%type	<tv_cfd>	rangecost
%type	<tv_char>	Class
%type	<tv_char>	Sentense
%type	<tv_chars>	Words

/*
 * Reserved words
 */
%token		EQ
%token		LB
%token		RB
%token		COMMA
%token		COLON
%token		SEMICOLON
%token		PERCENT	/* '%' */
%token		ALL	/* '*' */
%token		DASH	/* '-' */
%token		C_HOUR
%token		C_DAY
%token		C_YEAR
%token		C_CLASS
%token		C_DEFAULT
%token		C_PER
%token		C_COMMENT
%token		C_PREF

%%
cfg_file:
	| sequence;

sequence: seq ps
	| sequence seq ps
	;

ps: | SEMICOLON;

EatWord:
	| Sentense

EatEQ:
	| EQ

seq:
	rangecost {
		$1.year1=years.from;
		$1.year2=years.to;
		$1.cls=cls?strdup(cls):NULL;
		if(!$1.per) $1.per=1;
		$1.specific = 0;
		if($1.year1 != $1.year2)	$1.specific|=0x1000000;
		if($1.year2 - $1.year1 == 1)	$1.specific|=0x2000000;
		if($1.month1 != $1.month2)	$1.specific|=0x10000;
		if($1.month2 - $1.month1 == 1)	$1.specific|=0x20000;
		if($1.day1 != $1.day2)		$1.specific|=0x100;
		if($1.day2 - $1.day1 == 1)	$1.specific|=0x200;
		if($1.wday)	$1.specific|=0x400;
		if($1.hour1 != $1.hour2)	$1.specific|=0x1;
		if($1.hour2 - $1.hour1 == 1)	$1.specific|=0x2;

		lc_cfdata=(lc_t_cfdata **)_lc_realloc(lc_cfdata,
			sizeof(lc_t_cfdata *) * (++lc_cfdsize + 1));
		lc_cfdata[lc_cfdsize-1]=(lc_t_cfdata *)_lc_malloc(sizeof(lc_t_cfdata));
		lc_cfdata[lc_cfdsize-1][0]=$1;
		lc_cfdata[lc_cfdsize]=0;
	}
	| YearRange { years=$1; }
	| C_CLASS C_PREF EatEQ Words {
		if(lc_cpref) {
			char **z;
			for(z = lc_cpref; *z; z++)
				free(*z);
			free(lc_cpref);
		};
		lc_cpref=$4;
	}
	| Class {
		if(cls) free(cls);
		cls=$1;
	}
	;

Sentense:
	TOK_STRING {
		$$=$1;
	}
	| Sentense TOK_STRING {
		$$=(char *)_lc_realloc($1, strlen($1) + strlen($2) + 2);
		strcat($$, " ");
		strcat($$, $2);
	};

Comment:	C_COMMENT;

rangecost:
	Period Cost Comment Sentense {
		memcpy(&($$), &($1), sizeof(lc_cfdata));
		$$.type = $2.type;
		$$.per = $2.per;
		$$.price = $2.price;
		$$.comment=$4;
	}
	| Period Cost {
		memcpy(&($$), &($1), sizeof(lc_cfdata));
		$$.type = $2.type;
		$$.per = $2.per;
		$$.price = $2.price;
	}
	;

Cost:
	TOK_STRING EatWord TOK_INT PERCENT {
		memset(&($$), 0, sizeof($$));
		$$.type = $1;
		$$.price = (double)$3 / 100;
		$$.per = 1;
	}
	| TOK_STRING EatWord TOK_INT EatWord C_PER TOK_INT EatWord {
		memset(&($$), 0, sizeof($$));
		$$.type = $1;
		$$.price = $3;
		$$.per = $6;
	}
	| TOK_STRING EatWord TOK_DOUBLE EatWord C_PER TOK_INT EatWord {
		memset(&($$), 0, sizeof($$));
		$$.type = $1;
		$$.price = $3;
		$$.per = $6;
	}
	;

Period:
	HourPeriod {
		$$=$1;
	}
	| DayPeriod {
		$$=$1;
	}
	| DayPeriod HourPeriod {
		$$=$1;
		$$.hour1=$2.hour1;
		$$.hour2=$2.hour2;
	}
	;

YearRange:
	C_YEAR NumericRange {
		if($2.from > 1900) $2.from-=1900;
		if($2.to > 1900) $2.to-=1900;
		if($2.from < 50) $2.from+=100;
		if($2.to < 50) $2.to+=100;
		$$=$2;
	}
	;

Class:
	C_CLASS TOK_STRING { $$=$2; }
	| C_CLASS C_DEFAULT { $$=NULL; }
	;

DayPeriod:
	C_DAY NumericRange {
		memset(&($$), 0, sizeof($$));
		$$.day1 = $2.from;
		$$.day2 = $2.to;
	}
	| C_DAY TOK_MONTH NumericRange {
		memset(&($$), 0, sizeof($$));
		$$.month2 = ($$.month1 = $2) + 1;
		$$.day1 = $3.from;
		$$.day2 = $3.to;
	}
	| C_DAY TOK_MONTH DASH TOK_MONTH NumericRange {
		memset(&($$), 0, sizeof($$));
		$$.month1 = $2;
		$$.month2 = $4;
		$$.day1 = $5.from;
		$$.day2 = $5.to;
	}
	| C_DAY TOK_WDAY {
		memset(&($$), 0, sizeof($$));
		$$.wday = $2;
	}
	| C_DAY TOK_MONTH TOK_WDAY {
		memset(&($$), 0, sizeof($$));
		$$.month2 = ($$.month1 = $2) + 1;
		$$.wday = $3;
	}
	| C_DAY TOK_WDAY TOK_MONTH {
		memset(&($$), 0, sizeof($$));
		$$.month2 = ($$.month1 = $3) + 1;
		$$.wday = $2;
	}
	;

HourPeriod:
	C_HOUR HourRange {
		memset(&($$), 0, sizeof($$));
		$$.hour1 = $2.from;
		$$.hour2 = $2.to;
	}
	;

HourRange:
	NumericRange {
		if($1.from < 0) $1.from = 0;
		if($1.from > 23) $1.from = 23;
		if($1.to > 24) $1.to = 24;
		if($1.to < 1) $1.to = 1;
		$$=$1;
	}
	;

NumericRange:
	TOK_INT { $$.from = $1; $$.to = $1+1; }
	| TOK_INT DASH TOK_INT { $$.from = $1; $$.to = $3; }
	| DASH TOK_INT { $$.from = 0; $$.to = $2; }
	| TOK_INT DASH { $$.from = $1; $$.to = 100; }
	| ALL { $$.from=0; $$.to=0; };
	;

Words:
	TOK_STRING {
		$$=(char **)_lc_calloc(2, sizeof(char *));
		if(!$$)
			abort();
		*($$)=$1;
	}
	| Words TOK_STRING {
		$$=(char **)_lc_realloc($1, lc_i_cc($1) + 1);
		$$[lc_i_cc($$)]=$2;
		$$[lc_i_cc($$)]=NULL;
	};

%%

int yyerror(char *s) {
	fprintf(stderr, "Config parse error near line %d: %s\n", lc_cfg_i_nline+1, s);
	return 0;
};

int lc_i_cc(char **c) {
	register int z=0;
	if(!c) return 0;
	while(*c++) z++;
	return z;
};
