#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include "cost.h"
#include "intern.h"

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

lc_t_cfdata **lc_cfdata=NULL;
int lc_cfdsize=0;

unsigned int _lc_f_maxch=0;
unsigned int _lc_f_donttoch=0;

char **lc_cpref=NULL;

int lc_error=0;

double _lc_i_cost(struct tm *, double spent);
lc_t_cfdata *_lc_i_findcfg(struct tm *tm); /* Find in checked */
int _lc_i_within(char **, char *);
int _lc_i_check(char *Type, char **Classes);

double _lc_cost_char(void *, size_t, size_t, uint, uint, int, char *, char **);
double _lc_cost_short(void *, size_t, size_t, uint, uint, int, char *, char **);
double _lc_cost_int(void *, size_t, size_t, uint, uint, int, char *, char **);
double _lc_cost_long(void *, size_t, size_t, uint, uint, int, char *, char **);

double
lc_timecost(Start, Elapsed, Type, Classes)
	time_t	Start;
	uint	Elapsed;
	char	*Type;
	char	**Classes;
{
	struct tm tm;
	double cost = 0;
	int diff;

	lc_error=0;

	/* Check all necessary blocks */
	_lc_i_check(Type, Classes);
	localtime_r(&Start, &tm);

	diff = 60*(59-tm.tm_min) + (60-tm.tm_sec);
	if(diff >= Elapsed)
		return _lc_i_cost(&tm, Elapsed);

	Elapsed -= diff;
	cost = _lc_i_cost(&tm, diff);

	tm.tm_min=0;
	tm.tm_sec=0;
	
	while(Elapsed > 3600) {
		tm.tm_hour++;
		mktime(&tm);
		cost += _lc_i_cost(&tm, 3600);
		Elapsed-=3600;
	}

	cost += _lc_i_cost(&tm, Elapsed);

	return cost;
}


double
_lc_i_cost(tm, spent)
	struct	tm *tm;
	double	spent;
{
	register lc_t_cfdata *cfd;

	if(!(cfd=_lc_i_findcfg(tm))) {
		lc_error=1;
		return 0;
	}

	if(cfd->per <= 1)	/* Avoid division by strange numbers */
		return (spent * cfd->price);

	return ((spent / cfd->per) * cfd->price);
}

lc_t_cfdata *_lc_i_findcfg(struct tm *tm) {
	lc_t_cfdata **cfdp;
	register lc_t_cfdata *cfd;
	register int i;

	if(tm == NULL) {
		errno = EINVAL;
		return NULL;
	}

	/* Scanning through marked lines of configuration */
	for(i=_lc_f_maxch; i >= 1; i--) {
		for(cfdp = lc_cfdata; (cfd=*cfdp); cfdp++) {
			if(cfd->f_lp != i) continue;

			if(cfd->year1 &&
				(tm->tm_year < cfd->year1 ||
				tm->tm_year >= cfd->year2)) continue;

			if(cfd->month1 &&
				((tm->tm_mon+1) < cfd->month1 ||
				(tm->tm_mon+1) >= cfd->month2)) continue;

			if(cfd->day1 &&
				(tm->tm_mday < cfd->day1 ||
				tm->tm_mday >= cfd->day2 )) continue;

			if(cfd->wday && tm->tm_wday != (cfd->wday-1)) continue;

			if(cfd->hour1 || cfd->hour2) {
				if(cfd->hour1 < cfd->hour2) {
					if(tm->tm_hour < cfd->hour1 ||
					tm->tm_hour >= cfd->hour2) continue;
				} else {
					if(tm->tm_hour < cfd->hour1 &&
					tm->tm_hour >= cfd->hour2) continue;
				}
			}

			return cfd;
		}
	}

	return NULL;
}

int
_lc_i_within(char **a, char *b) {
	register int z;
	register char **vp = a;

	if(!vp || !b) return 0;

	for(z=1; *vp; vp++) {
		if(!strcasecmp(*vp, b)) {
			while(*++vp)
				z++;
			return z;
		}
	}
	return 0;
}

int _lc_i_check(char *Type, char **Classes) {
	register lc_t_cfdata **cfdp;
	register lc_t_cfdata *cfd;
	_lc_f_maxch = 1;

	if(_lc_f_donttoch) return _lc_f_maxch;

	for(cfdp = lc_cfdata; (cfd = *cfdp); cfdp++) {

		/* Types are different */
		if(!cfd->type || strcasecmp(cfd->type, Type))
			{ cfd->f_lp=0; continue; }

		/* If class is absent, this is a common rule */
		if(!cfd->cls) { cfd->f_lp=1; continue; }

		/* If class present, but not mentioned when called */
		if(cfd->cls && !Classes) { cfd->f_lp=0; continue; }

		/* Classes are different */
		if(!_lc_i_within(Classes, cfd->cls)) { cfd->f_lp=0; continue; }

		/* Shifting back things needs to be processed later */
		cfd->f_lp= 1 + _lc_i_within(lc_cpref, cfd->cls);

		if(cfd->f_lp > _lc_f_maxch)
			_lc_f_maxch = cfd->f_lp;

	}

	return _lc_f_maxch;

}

double
lc_cost(base, nmemb, size, timeshift, datashift, datasize, vit, Type, Classes)
	void	*base;
	size_t	nmemb;
	size_t	size;
	uint	timeshift;
	uint	datashift;
	size_t	datasize;
	int		vit;
	char	*Type;
	char	**Classes;
{
	/* Fool-proofing */
	lc_error=1;
	_lc_i_check(Type, Classes);
	if(!base) {
		errno = EINVAL;
		return 0;
	}
	if(nmemb <= 0) {
		errno = EINVAL;
		return 0;
	}
	if(timeshift >= size + sizeof(time_t)) {
		return 0;
	}

	if(!datasize) datasize = sizeof(unsigned int);
	if(datashift >= size + datasize) return 0;

	lc_error=0;

	switch(datasize) {
	case 1:
		return _lc_cost_char(base, nmemb, size,
			timeshift, datashift, vit, Type, Classes);
	case 2:
		return _lc_cost_short(base, nmemb, size,
			timeshift, datashift, vit, Type, Classes);
	case 4:
		return _lc_cost_int(base, nmemb, size,
			timeshift, datashift, vit, Type, Classes);
	case 8:
		return _lc_cost_long(base, nmemb, size,
			timeshift, datashift, vit, Type, Classes);
	default:
		break;
	}

	lc_error=1;
	return 0;
}

double
_lc_cost_char(base, nmemb, size, timeshift, datashift, vit, Type, Classes)
	void	*base;
	size_t	nmemb;
	size_t	size;
	uint	timeshift;
	uint	datashift;
	int		vit;
	char	*Type;
	char	**Classes;
{
	double cost=0;
	if(vit) {
		_lc_f_donttoch=1;
		while(nmemb--) {
			cost+=lc_timecost(*(time_t *)((char *)base + timeshift),
			*((char *)base + datashift), Type, Classes);
			(char *)base += size; 
		}
		_lc_f_donttoch=0;
	} else {
		struct tm tm;
		while(nmemb--) {
			localtime_r((time_t *)((char *)base + timeshift), &tm);
			cost+=_lc_i_cost(&tm, *((char *)base + datashift));
			(char *)base += size; 
		}
	}
	return cost;
}

double _lc_cost_short(base, nmemb, size, timeshift, datashift, vit, Type, Classes)
	void	*base;
	size_t	nmemb;
	size_t	size;
	uint	timeshift;
	uint	datashift;
	int		vit;
	char	*Type;
	char	**Classes;
{
	double cost=0;
	if(vit) {
		_lc_f_donttoch=1;
		while(nmemb--) {
			cost+=lc_timecost(*(time_t *)((char *)base + timeshift),
			*(short *)((char *)base + datashift), Type, Classes);
			(char *)base += size; 
		}
		_lc_f_donttoch=0;
	} else {
		struct tm tm;
		while(nmemb--) {
			localtime_r((time_t *)((char *)base + timeshift), &tm);
			cost+=_lc_i_cost(&tm, *(short *)((char *)base + datashift));
			(char *)base += size; 
		}
	}
	return cost;
}


double _lc_cost_int(base, nmemb, size, timeshift, datashift, vit, Type, Classes)
	void	*base;
	size_t	nmemb;
	size_t	size;
	uint	timeshift;
	uint	datashift;
	int		vit;
	char	*Type;
	char	**Classes;
{
	double cost=0;
	if(vit) {
		_lc_f_donttoch=1;
		while(nmemb--) {
			cost+=lc_timecost(*(time_t *)((char *)base + timeshift),
			*(int *)((char *)base + datashift), Type, Classes);
			(char *)base += size; 
		}
		_lc_f_donttoch=0;
	} else {
		struct tm tm;
		while(nmemb--) {
			localtime_r((time_t *)((char *)base + timeshift), &tm);
			cost+=_lc_i_cost(&tm, *(int *)((char *)base + datashift));
			(char *)base += size; 
		}
	}
	return cost;
}


double
_lc_cost_long(base, nmemb, size, timeshift, datashift, vit, Type, Classes)
	void	*base;
	size_t	nmemb;
	size_t	size;
	uint	timeshift;
	uint	datashift;
	int		vit;
	char	*Type;
	char	**Classes;
{
	double cost=0;
	if(vit) {
		_lc_f_donttoch=1;
		while(nmemb--) {
			cost += lc_timecost(
				*(time_t *)((char *)base + timeshift),
				*(long long *)((char *)base + datashift),
				Type, Classes);
			(char *)base += size; 
		}
		_lc_f_donttoch=0;
	} else {
		struct tm tm;
		while(nmemb--) {
			localtime_r((time_t *)((char *)base + timeshift), &tm);
			cost += _lc_i_cost(&tm,
				*(long long *)((char *)base + datashift));
			(char *)base += size; 
		}
	}
	return cost;
}

