#ifndef __COST_H__
#define __COST_H__

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Process configuration file and initialize libcost.
 */
int lc_init(const char *cfgname);

/*
 * Close session and free associated memory
 */
void lc_end(void);

extern int lc_error;

/*
 * Count the cost of time.
 */
double
lc_timecost(
	time_t	Start,
	uint	Elapsed,
	char	*Type,
	char	**optionalClasses
);


/*
 * Count cost of any service. Base is the start address of the
 * array of nmemb elements each size long. Timeshift points to the
 * time_t field of each element. Datashift points to a char,
 * short, int, or long long field. Field type guessed by the datasize arg.
 *
 * For example:
 * We have the following structure:
 *
 * struct data {
 *   unsigned int flags;
 *   time_t time;
 *   unsigned int value;
 * };
 * struct data *datap;
 *
 * Then we should call lc_cost in the following manner:
 * 
 * cost = lc_cost(datap, nmemb, sizeof(struct data),
 *   sizeof(flags),	// Time shift
 *   sizeof(time_t) + sizeof(flags),	// Data shift
 *   sizeof(unsigned int),	// Data size
 *   0, // if 1, value will be considered as time spend since time_t field
 *   "online",	// char *Type
 *   NULL);
 *
 */

double
lc_cost(
	void	*base,
	size_t	nmemb,
	size_t	size,
	uint	timeshift,
	uint	datashift,
	size_t	datasize,
	int		valueistime,
	char *Type,
	char **optionalClasses
);


#ifdef __cplusplus
}
#endif

#endif
